#!/usr/bin/env perl
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#

unshift @INC, ( '.' );

use AI::FreeHAL::Engine;
use AI::Util;
use AI::SemanticNetwork;
use AI::POS;
use AI::Selector;
use AI::Tagger;

1;